from datetime import datetime
from enum import Enum
from itertools import chain
from typing import Dict, List, Union

from dateutil import tz

Comparable = Union[str, Enum, datetime, bool]


def _format_kwargs(kwargs: Dict) -> List[str]:
    return [f"{k} {v}" for k, v in kwargs.items()]


def _format_comparable(value: Comparable) -> str:
    if isinstance(value, Enum):
        return f"'{value.value}'"
    elif isinstance(value, datetime):
        return value.astimezone(tz.UTC).strftime("%Y-%m-%dT%H:%M:%SZ")
    elif value is True:
        return "true"
    elif value is False:
        return "false"
    else:
        return f"'{value}'"


def _generate_operator(op, args, kwargs) -> str:
    if len(args) + len(kwargs) == 1:
        return "".join(chain(args, _format_kwargs(kwargs)))
    condition = f" {op} ".join(chain(args, _format_kwargs(kwargs)))
    return f"({condition.strip()})"


def and_(*args, **kwargs) -> str:
    return _generate_operator("and", args, kwargs)


def or_(*args, **kwargs) -> str:
    return _generate_operator("or", args, kwargs)


def eq(value: Comparable) -> str:
    return f"eq {_format_comparable(value)}"


def ne(value: Comparable) -> str:
    return f"ne {_format_comparable(value)}"


def gt(value: Comparable) -> str:
    return f"gt {_format_comparable(value)}"


def ge(value: Comparable) -> str:
    return f"ge {_format_comparable(value)}"


def lt(value: Comparable) -> str:
    return f"lt {_format_comparable(value)}"


def le(value: Comparable) -> str:
    return f"le {_format_comparable(value)}"



from abc import ABC

from .converters import BaseConverter


class BaseConvertableObject(ABC):
    ATTRIBUTES = ()

    def to_dict(self):
        """Convert resource into dict accepted by API"""
        data_dict = {}

        for attr in self.ATTRIBUTES:
            if isinstance(attr, BaseConverter):
                value = getattr(self, attr.local_name, None)
                data_dict[attr.original_name] = attr.back_converter(value)
            else:
                data_dict[attr] = getattr(self, attr, None)

        return data_dict

    @classmethod
    def create_from_dict(cls, data_dict: dict, **additional_kwargs):
        init_arguments = {}
        private_attributes = {}

        def store_attribute(name, value):
            if name.startswith("_"):
                private_attributes[name] = value
            else:
                init_arguments[name] = value

        for attr in cls.ATTRIBUTES:
            if isinstance(attr, BaseConverter):
                if attr.original_name in data_dict:
                    value = attr.obj_converter(data_dict.get(attr.original_name))
                    store_attribute(attr.local_name, value)
            elif attr in data_dict:
                store_attribute(attr, data_dict.get(attr))

        obj = cls(**init_arguments, **additional_kwargs)
        for attr, value in private_attributes.items():
            setattr(obj, attr, value)

        return obj



from dataclasses import dataclass

from .patterns import BaseRecurrencePattern
from .ranges import BaseRecurrenceRange


@dataclass
class Recurrence:
    pattern: BaseRecurrencePattern
    range: BaseRecurrenceRange

    def to_dict(self):
        return {"pattern": self.pattern.to_dict(), "range": self.range.to_dict()}



from datetime import datetime

from ..attributes import RecurrenceRangeType
from ..convertable import BaseConvertableObject
from ..converters.basic import (
    AttributeConverter,
    DateAttrConverter,
    RecurrenceRangeTypeAttrConverter,
)


class BaseRecurrenceRange(BaseConvertableObject):
    ATTRIBUTES = (
        RecurrenceRangeTypeAttrConverter("type", "_range_type"),
        DateAttrConverter("startDate", "start_date"),
    )

    def __init__(self, range_type: RecurrenceRangeType, start_date: datetime):
        self._range_type = range_type
        self.start_date = start_date


class EndDate(BaseRecurrenceRange):
    ATTRIBUTES = (
        *BaseRecurrenceRange.ATTRIBUTES,
        DateAttrConverter("endDate", "end_date"),
    )

    def __init__(self, start_date: datetime, end_date: datetime):
        super().__init__(RecurrenceRangeType.END_DATE, start_date)
        self.end_date = end_date


class NoEnd(BaseRecurrenceRange):
    def __init__(self, start_date: datetime):
        super().__init__(RecurrenceRangeType.NO_END, start_date)


class Numbered(BaseRecurrenceRange):
    ATTRIBUTES = (
        *BaseRecurrenceRange.ATTRIBUTES,
        AttributeConverter("numberOfOccurrences", "occurrences"),
    )

    def __init__(self, start_date: datetime, occurrences: int):
        super().__init__(RecurrenceRangeType.NUMBERED, start_date)
        self.occurrences = occurrences



from abc import ABC
from datetime import datetime

from furl import furl

from .attributes import Importance, Status
from .convertable import BaseConvertableObject
from .converters.basic import (
    AttributeConverter,
    ContentAttrConverter,
    DatetimeAttrConverter,
    ImportanceAttrConverter,
    IsoTimeAttrConverter,
    StatusAttrConverter,
)
from .converters.recurrence import RecurrenceAttrConverter
from .filters import and_, ne


class ResourceAlreadyCreatedError(Exception):
    """This resource is already created. Prevent duplicate"""


class TaskListNotSpecifiedError(Exception):
    """TaskList id must be set before create task"""


class Resource(BaseConvertableObject, ABC):
    """Base Resource for any other"""

    ENDPOINT = ""
    ATTRIBUTES = ()

    def __init__(self, client):
        self._client = client

    def create(self):
        """Create object in API"""
        if self.id:
            raise ResourceAlreadyCreatedError
        result = self._client.raw_post(self.managing_endpoint, self.to_dict(), 201)
        # TODO: update object from result
        self._id = result.get("id", None)

    def update(self):
        """Update resource in API"""
        self._client.patch(self)

    def delete(self):
        """Delete object in API"""
        self._client.delete(self)

    @property
    def managing_endpoint(self):
        return (furl(self.ENDPOINT) / (self.id or "")).url

    @property
    def id(self):
        return getattr(self, "_id", None)

    @classmethod
    def create_from_dict(cls, client, data_dict):
        return super().create_from_dict(data_dict, client=client)

    @classmethod
    def handle_list_filters(cls, *args, **kwargs):
        if len(args) + len(kwargs) == 0:
            return {}
        params = {"$filter": and_(*args, **kwargs)}
        return params


class TaskList(Resource):
    """Represent a list of tasks"""

    ENDPOINT = "todo/lists"
    ATTRIBUTES = (
        AttributeConverter("id", "_id"),
        AttributeConverter("displayName", "name"),
        AttributeConverter("isShared", "_is_shared"),
        AttributeConverter("isOwner", "_is_owner"),
        AttributeConverter("wellknownListName", "_well_known_name"),
    )

    def __init__(self, client, name: str):
        super().__init__(client)
        self.name = name

    def __repr__(self):
        return f"<TaskList '{self.name}'>"

    def __str__(self):
        return f"List '{self.name}'"

    def get_tasks(self, *args, **kwargs):
        """Get list of tasks in given list. Default returns only non-completed tasks."""
        tasks_endpoint = furl(self.ENDPOINT) / self.id / "tasks"
        return self._client.list(Task, endpoint=tasks_endpoint.url, *args, **kwargs)

    def save_task(self, task):
        task.task_list = self
        task.create()


class Task(Resource):
    """Represent a task."""

    ENDPOINT = "tasks"
    ATTRIBUTES = (
        AttributeConverter("id", "_id"),
        ContentAttrConverter("body", "body"),
        StatusAttrConverter("status", "status"),
        "title",
        RecurrenceAttrConverter("recurrence", "recurrence"),
        ImportanceAttrConverter("importance", "importance"),
        AttributeConverter("isReminderOn", "is_reminder_on"),
        IsoTimeAttrConverter("createdDateTime", "created_datetime"),
        DatetimeAttrConverter("dueDateTime", "due_datetime"),
        DatetimeAttrConverter("completedDateTime", "completed_datetime"),
        IsoTimeAttrConverter("lastModifiedDateTime", "last_modified_datetime"),
        DatetimeAttrConverter("reminderDateTime", "reminder_datetime"),
    )

    def __init__(
        self,
        client,
        title: str,
        body: str = None,
        status: Status = None,
        importance: Importance = None,
        recurrence: dict = None,
        is_reminder_on: bool = False,
        created_datetime: datetime = None,
        due_datetime: datetime = None,
        completed_datetime: datetime = None,
        last_modified_datetime: datetime = None,
        reminder_datetime: datetime = None,
        task_list: TaskList = None,
    ):
        super().__init__(client)
        self.body = body
        self.title = title
        self.status = status
        self.importance = importance
        self.recurrence = recurrence
        self.is_reminder_on = is_reminder_on
        self.created_datetime = created_datetime
        self.due_datetime = due_datetime
        self.completed_datetime = completed_datetime
        self.last_modified_datetime = last_modified_datetime
        self.reminder_datetime = reminder_datetime
        self.task_list = task_list

    def __repr__(self):
        return f"<Task '{self.title}'>"

    def __str__(self):
        return f"Task '{self.title}'"

    def create(self):
        if not self.task_list:
            raise TaskListNotSpecifiedError
        return super().create()

    @classmethod
    def handle_list_filters(cls, *args, **kwargs):
        kwargs.setdefault("status", ne(Status.COMPLETED))
        return super().handle_list_filters(*args, **kwargs)

    @property
    def managing_endpoint(self):
        return (furl(self.task_list.managing_endpoint) / super().managing_endpoint).url



from typing import Iterable

from ..attributes import RecurrencePatternType, Weekday
from ..convertable import BaseConvertableObject
from ..converters.basic import (
    AttributeConverter,
    ListConverter,
    RecurrencePatternTypeAttrConverter,
    WeekdayAttrConverter,
)


class BaseRecurrencePattern(BaseConvertableObject):
    ATTRIBUTES = (
        "interval",
        RecurrencePatternTypeAttrConverter("type", "_pattern_type"),
    )

    def __init__(self, pattern_type: RecurrencePatternType, interval: int):
        self._pattern_type = pattern_type
        self.interval = interval


class Daily(BaseRecurrencePattern):
    def __init__(self, interval: int):
        super().__init__(RecurrencePatternType.DAILY, interval)


class Weekly(BaseRecurrencePattern):
    ATTRIBUTES = (
        *BaseRecurrencePattern.ATTRIBUTES,
        WeekdayAttrConverter("firstDayOfWeek", "week_start"),
        ListConverter(WeekdayAttrConverter("daysOfWeek", "days_of_week")),
    )

    def __init__(
        self,
        interval: int,
        days_of_week: Iterable[Weekday],
        week_start: Weekday = Weekday.SUNDAY,
    ):
        super().__init__(RecurrencePatternType.WEEKLY, interval)
        self.days_of_week = days_of_week
        self.week_start = week_start


class MonthlyAbsolute(BaseRecurrencePattern):
    ATTRIBUTES = (
        *BaseRecurrencePattern.ATTRIBUTES,
        AttributeConverter("dayOfMonth", "day_of_month"),
    )

    def __init__(self, interval: int, day_of_month: int):
        super().__init__(RecurrencePatternType.MONTHLY_ABSOLUTE, interval)
        self.day_of_month = day_of_month


class MonthlyRelative(BaseRecurrencePattern):
    ATTRIBUTES = (
        *BaseRecurrencePattern.ATTRIBUTES,
        ListConverter(WeekdayAttrConverter("daysOfWeek", "days_of_week")),
    )

    def __init__(self, interval: int, days_of_week: Iterable[Weekday]):
        super().__init__(RecurrencePatternType.MONTHLY_RELATIVE, interval)
        self.days_of_week = days_of_week


class YearlyAbsolute(BaseRecurrencePattern):
    ATTRIBUTES = (
        *BaseRecurrencePattern.ATTRIBUTES,
        AttributeConverter("dayOfMonth", "day_of_month"),
        "month",
    )

    def __init__(self, interval: int, day_of_month: int, month: int):
        super().__init__(RecurrencePatternType.YEARLY_ABSOLUTE, interval)
        self.day_of_month = day_of_month
        self.month = month


class YearlyRelative(BaseRecurrencePattern):
    ATTRIBUTES = (
        *BaseRecurrencePattern.ATTRIBUTES,
        ListConverter(WeekdayAttrConverter("daysOfWeek", "days_of_week")),
        "month",
    )

    def __init__(self, interval: int, days_of_week: Iterable[Weekday], month: int):
        super().__init__(RecurrencePatternType.YEARLY_RELATIVE, interval)
        self.days_of_week = days_of_week
        self.month = month



from abc import ABC
from dataclasses import dataclass
from datetime import date, datetime
from enum import Enum
from typing import Any, List

from dateutil import parser, tz

from ..attributes import (
    Importance,
    RecurrencePatternType,
    RecurrenceRangeType,
    Status,
    Weekday,
)
from . import BaseConverter


@dataclass
class AttributeConverter(BaseConverter):
    original_name: str
    local_name: str

    def obj_converter(self, data: Any) -> Any:
        return data

    def back_converter(self, data: Any) -> Any:
        return data


@dataclass
class DatetimeAttrConverter(AttributeConverter):
    def obj_converter(self, data: dict) -> datetime:
        if not data:
            return None

        date = parser.parse(data["dateTime"])
        return datetime.combine(date.date(), date.time(), tz.gettz(data["timeZone"]))

    def back_converter(self, data: datetime) -> dict:
        if not data:
            return None

        return {
            "dateTime": data.strftime("%Y-%m-%dT%H:%M:%S.%f"),
            "timeZone": data.tzname(),
        }


@dataclass
class ContentAttrConverter(AttributeConverter):
    def obj_converter(self, data: dict) -> str:
        if not data:
            return ""
        return data["content"]

    def back_converter(self, data: str) -> dict:
        return {"content": data, "contentType": "html"}


@dataclass
class IsoTimeAttrConverter(AttributeConverter):
    def obj_converter(self, data: str) -> datetime:
        return parser.isoparse(data)

    def back_converter(self, data: datetime) -> str:
        if not data:
            return None
        return data.astimezone(tz.UTC).strftime("%Y-%m-%dT%H:%M:%SZ")


@dataclass
class DateAttrConverter(AttributeConverter):
    def obj_converter(self, data: str) -> date:
        return parser.parse(data).date()

    def back_converter(self, data: date) -> str:
        if not data:
            return None
        return data.isoformat()


class ListConverter(AttributeConverter):
    def __init__(self, obj_converter: AttributeConverter):
        self._converter = obj_converter

    def obj_converter(self, data: List[Any]) -> List[Any]:
        if data is None:
            return None
        return [self._converter.obj_converter(element) for element in data]

    def back_converter(self, data: List[Any]) -> List[Any]:
        if data is None:
            return None
        return [self._converter.back_converter(element) for element in data]

    @property
    def original_name(self) -> str:
        return self._converter.original_name

    @property
    def local_name(self) -> str:
        return self._converter.local_name


@dataclass
class EnumAttrConverter(AttributeConverter, ABC):
    _ENUM = None

    def obj_converter(self, data: str) -> Enum:
        if not data:
            return None
        return self._ENUM(data)

    def back_converter(self, data: Enum) -> str:
        if not data:
            return None
        return data.value


@dataclass
class ImportanceAttrConverter(EnumAttrConverter):
    _ENUM = Importance


@dataclass
class StatusAttrConverter(EnumAttrConverter):
    _ENUM = Status


@dataclass
class RecurrencePatternTypeAttrConverter(EnumAttrConverter):
    _ENUM = RecurrencePatternType


@dataclass
class RecurrenceRangeTypeAttrConverter(EnumAttrConverter):
    _ENUM = RecurrenceRangeType


@dataclass
class WeekdayAttrConverter(EnumAttrConverter):
    _ENUM = Weekday



from typing import Type

from furl import furl
from requests import Response, codes

from .provider import AbstractProvider
from .resources import Resource


class ToDoClient(object):
    def __init__(
        self,
        provider: AbstractProvider,
        api_url: str = "https://graph.microsoft.com/beta",
        api_prefix: str = "me",
    ):
        self._provider = provider
        self._url = furl(api_url) / api_prefix

    def list(self, resource_class: Type[Resource], endpoint: str = None, **kwargs):
        url = (self._url / (endpoint or resource_class.ENDPOINT)).url
        params = resource_class.handle_list_filters(**kwargs)

        elements = []
        while url:
            response = self._provider.get(url, params=params)
            self._map_http_errors(response, codes.ok)
            data = response.json()
            elements += data["value"]
            url = data.get("@odata.nextLink", None)
            params = {}

        return [resource_class.create_from_dict(self, element) for element in elements]

    def get(self, resource_class: Type[Resource], resource_id: str):
        url = (self._url / resource_class.ENDPOINT / resource_id).url
        response = self._provider.get(url)
        self._map_http_errors(response, codes.ok)
        return resource_class.create_from_dict(self, response.json())

    def delete(self, resource: Resource):
        url = (self._url / resource.managing_endpoint).url
        response = self._provider.delete(url)
        self._map_http_errors(response, codes.no_content)

    def patch(self, resource: Resource):
        url = (self._url / resource.managing_endpoint).url
        response = self._provider.patch(url, json_data=resource.to_dict())
        self._map_http_errors(response, codes.ok)
        return response.json()

    def raw_post(self, endpoint: str, data: dict, expected_code: int = codes.created):
        url = (self._url / endpoint).url
        response = self._provider.post(url, json_data=data)
        self._map_http_errors(response, expected_code)
        return response.json()

    def _map_http_errors(self, response, expected):
        if response.status_code == codes.not_found:
            raise ResourceNotFoundError(response)

        if response.status_code != expected:
            raise ResponseError(response)


class ResponseError(Exception):
    """Response returned an error"""

    MESSAGE = None

    def __init__(self, response: Response):
        self.response = response

    def __str__(self):
        if self.MESSAGE:
            return self.MESSAGE

        details = f"{self.response.status_code} {self.response.reason}"
        return f"Server returned an error: {details}"


class ResourceNotFoundError(ResponseError):
    """Requested resource not exists"""

    MESSAGE = "404 Resource not found"



from enum import Enum


class Importance(Enum):
    LOW = "low"
    NORMAL = "normal"
    HIGH = "high"


class Status(Enum):
    NOT_STARTED = "notStarted"
    IN_PROGRESS = "inProgress"
    COMPLETED = "completed"
    WAITING_ON_OTHERS = "waitingOnOthers"
    DEFERRED = "deferred"


class RecurrencePatternType(Enum):
    DAILY = "daily"
    WEEKLY = "weekly"
    MONTHLY_ABSOLUTE = "absoluteMonthly"
    MONTHLY_RELATIVE = "relativeMonthly"
    YEARLY_ABSOLUTE = "absoluteYearly"
    YEARLY_RELATIVE = "relativeYearly"


class RecurrenceRangeType(Enum):
    END_DATE = "endDate"
    NO_END = "noEnd"
    NUMBERED = "numbered"


class Weekday(Enum):
    SUNDAY = "sunday"
    MONDAY = "monday"
    TUESDAY = "tuesday"
    WEDNESDAY = "wednesday"
    THURSDAY = "thursday"
    FRIDAY = "friday"
    SATURDAY = "saturday"



from ..attributes import RecurrencePatternType, RecurrenceRangeType
from ..recurrence import Recurrence, patterns, ranges
from .basic import AttributeConverter


class RecurrencePatternAttrConverter(AttributeConverter):
    _CONVERTING_TABLE = {
        RecurrencePatternType.DAILY.value: patterns.Daily,
        RecurrencePatternType.WEEKLY.value: patterns.Weekly,
        RecurrencePatternType.MONTHLY_ABSOLUTE.value: patterns.MonthlyAbsolute,
        RecurrencePatternType.MONTHLY_RELATIVE.value: patterns.MonthlyRelative,
        RecurrencePatternType.YEARLY_ABSOLUTE.value: patterns.YearlyAbsolute,
        RecurrencePatternType.YEARLY_RELATIVE.value: patterns.YearlyRelative,
    }

    def obj_converter(self, data: dict) -> patterns.BaseRecurrencePattern:
        if not data:
            return None

        pattern_class = self._CONVERTING_TABLE.get(data.get("type"))
        if not pattern_class:
            raise ValueError

        return pattern_class.create_from_dict(data)

    def back_converter(self, data: patterns.BaseRecurrencePattern) -> dict:
        if not data:
            return None
        return data.to_dict()


class RecurrenceRangeAttrConverter(AttributeConverter):
    _CONVERTING_TABLE = {
        RecurrenceRangeType.END_DATE.value: ranges.EndDate,
        RecurrenceRangeType.NO_END.value: ranges.NoEnd,
        RecurrenceRangeType.NUMBERED.value: ranges.Numbered,
    }

    def obj_converter(self, data: dict) -> ranges.BaseRecurrenceRange:
        if not data:
            return None

        pattern_class = self._CONVERTING_TABLE.get(data.get("type"))
        if not pattern_class:
            raise ValueError

        return pattern_class.create_from_dict(data)

    def back_converter(self, data: ranges.BaseRecurrenceRange) -> dict:
        if not data:
            return None
        return data.to_dict()


class RecurrenceAttrConverter(AttributeConverter):
    _range_converter = RecurrenceRangeAttrConverter("", "")
    _pattern_converter = RecurrencePatternAttrConverter("", "")

    def obj_converter(self, data: dict) -> Recurrence:
        if not data:
            return None

        return Recurrence(
            self._pattern_converter.obj_converter(data.get("pattern")),
            self._range_converter.obj_converter(data.get("range")),
        )

    def back_converter(self, data: Recurrence) -> dict:
        if not data:
            return None

        return data.to_dict()



import webbrowser
import wsgiref.simple_server
import wsgiref.util

from furl import furl
from requests_oauthlib import OAuth2Session

from .base import AbstractProvider


class WebBrowserProvider(AbstractProvider):
    """An provider that can call webbrowser to open sign-in page"""

    _SCOPES = "profile openid User.Read Calendars.Read Tasks.ReadWrite"
    _DEFAULT_OPEN_MESSAGE = (
        "Open following page in your webbrowser and finish singing in:"
    )
    _DEFAULT_FINISH_MESSAGE = (
        "Authorization completed. You can close this page and return to the app"
    )

    def __init__(
        self,
        app_id: str,
        app_secret: str,
        authority: str = "https://login.microsoftonline.com/common/",
        authorize_endpoint: str = "oauth2/v2.0/authorize",
        token_endpoint: str = "oauth2/v2.0/token",
        open_message: str = _DEFAULT_OPEN_MESSAGE,
        finish_message: str = _DEFAULT_FINISH_MESSAGE,
    ):
        self._app_id = app_id
        self._app_secret = app_secret
        self._open_message = open_message
        self._finish_message = finish_message

        authority = furl(authority)
        self._authorize_url = (authority / authorize_endpoint).url
        self._token_url = (authority / token_endpoint).url

        self._session = None

    def authorize(self, local_port: int = 8888, print_message: bool = True):
        """Run authorization workflow. Call webbrowser login, get response and token

        'local_port' - on this port we wait for redirection from sing-in page.
        Address http://localhost:<local_port>/ must be allowed as redirect URL. """

        redirect_url = f"http://localhost:{local_port}"

        self._session = self._build_session(redirect_url)

        sign_in_url, state = self._session.authorization_url(
            self._authorize_url, prompt="login"
        )

        response_app = _LocalRedirectHandlingApp(self._finish_message)
        response_server = wsgiref.simple_server.make_server(
            "localhost", local_port, response_app
        )

        if print_message:
            print(self._open_message)
            print(sign_in_url)

        webbrowser.open(sign_in_url)
        response_server.handle_request()

        self._token = self._session.fetch_token(
            self._token_url,
            client_secret=self._app_secret,
            authorization_response=self._replace_http_into_https(
                response_app.callback_url
            ),
        )

    def get(self, url: str, params: dict = None):
        if not self._token:
            raise RequestBeforeAuthenticatedError

        return self._session.get(url=url, params=params)

    def delete(self, url: str):
        if not self._token:
            raise RequestBeforeAuthenticatedError

        return self._session.delete(url=url)

    def patch(self, url: str, json_data: dict):
        if not self._token:
            raise RequestBeforeAuthenticatedError

        return self._session.patch(url=url, json=json_data)

    def post(self, url: str, json_data: dict):
        if not self._token:
            raise RequestBeforeAuthenticatedError

        return self._session.post(url=url, json=json_data)

    def _build_session(self, redirect_url: str):
        refresh_params = {"client_id": self._app_id, "client_secret": self._app_secret}

        session = OAuth2Session(
            self._app_id,
            scope=self._SCOPES,
            redirect_uri=self._replace_http_into_https(redirect_url),
            auto_refresh_url=self._token_url,
            auto_refresh_kwargs=refresh_params,
            token_updater=self._save_token,
        )
        # Workaround for InsecureTransportError from OAuthLib for http://localhost
        session.redirect_uri = redirect_url

        return session

    def _save_token(self, token):
        self._token = token

    def _replace_http_into_https(self, url: str):
        """OAuthLib strictly expects HTTPS protocol, even for localhost"""
        if url.startswith("http://"):
            return url.replace("http", "https", 1)
        return url


class _LocalRedirectHandlingApp(object):
    """A WSGI app handles redirection from sign-in service."""

    def __init__(self, message):
        self.callback_url = None
        self._message = message

    def __call__(self, environ, start_response):
        start_response("200 OK", [("Content-type", "text/plain")])
        self.callback_url = wsgiref.util.request_uri(environ)
        return [self._message.encode("utf-8")]


class RequestBeforeAuthenticatedError(Exception):
    """Try to execute request before authenticate"""



from setuptools import find_packages, setup


def readme():
    with open("README.md", "r") as file:
        return file.read()


def requirements():
    with open("requirements.in") as file:
        return file.read().splitlines()


setup(
    name="todo-ms-client",
    version="0.0.4",
    description="Unofficial Python library for MS ToDo API",
    long_description=readme(),
    long_description_content_type="text/markdown",
    install_requires=requirements(),
    url="https://github.com/kam193/todo-ms-client",
    author="Kamil Mańkowski",
    author_email="kam193@wp.pl",
    license="MIT",
    packages=find_packages(exclude=["*tests*"]),
    zip_safe=False,
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



from abc import ABC, abstractmethod

from requests import Response


class AbstractProvider(ABC):
    @abstractmethod
    def get(self, url: str, params: dict = None) -> Response:
        pass

    @abstractmethod
    def delete(self, url: str) -> Response:
        pass

    @abstractmethod
    def patch(self, url: str, json_data: dict) -> Response:
        pass

    @abstractmethod
    def post(self, url: str, json_data: dict) -> Response:
        pass
